'''
	Simple python3 script to get the code size of the AES encryption and 
	key schedule functions given at https://github.com/aadomn/aes.
	The output of arm-none-eabi-objdump has to be given as input parameter.
	If one function is not defined, the corresponding code size will be 0.

	!!! Be carefull !!! Some subroutines share the same name between the
	different implementations. Therefore, if you run this script to get the
	code size of e.g. fixsliced implementations, you should just ignore what is
	printed for barrel-shiftrows implementations (and vice-versa).

	author: Alexandre Adomnicai <alex.adomnicai@gmail.com>
	date: 	November 2020
'''
import sys

'''
Return the code size of a subroutine.
'''
def get_size_subroutine(filename, label):
	file = open(filename, 'r')
	start = 0
	end = 0
	line = file.readline()
	while line:
		if label in line: #as soon as we find the definition of the subroutine
			start = int(line.split()[0], base=16) #get its beginning address
			line = file.readline()
			while not(line == '\n'):
				line = file.readline()
			line = file.readline() #as soon as its definition ends
			end = int(line.split()[0], base=16) #get its ending address
			break
		line = file.readline()
	file.close()
	return end-start 

'''
Return the code size of a function, including the subroutines it uses.
'''
def get_size_function(filename, subroutines):
	total_size = 0
	for i in subroutines:
		total_size += get_size_subroutine(filename, i)
	return total_size

'''
AES-128 semi-fixsliced
'''
def get_size_aes128_keyschedule_sfs(filename):
	subroutines = ['<packing>:', '<aes128_xorcolumns_rotword>:',
		'<inv_shiftrows_1>:', '<loop_inv_sr_1>:', '<aes128_keyschedule_sfs>:']
	return get_size_function(filename, subroutines)

def get_size_aes128_keyschedule_sfs_lut(filename):
	subroutines = ['<packing_rkey>:', '<aes128_keyschedule_rfunc>:',
		'<inv_shiftrows_1>:', '<aes128_keyschedule_sfs_lut>:']
	return get_size_function(filename, subroutines)

def get_size_aes128_encrypt_sfs(filename):
	subroutines = ['<packing>:', '<unpacking>:', '<ark_sbox>:',
		'<mixcolumns_0>:', '<mixcolumns_3>:', '<double_shiftrows>:',
		'<aes128_encrypt_sfs>:']
	return get_size_function(filename, subroutines)

'''
AES-256 semi-fixsliced
'''
def get_size_aes256_keyschedule_sfs(filename):
	subroutines = ['<packing>:', '<aes256_xorcolumns_rotword>:',
		'<aes256_xorcolumns>:', '<inv_shiftrows_1>:', '<loop_inv_sr_1>:',
		'<aes256_keyschedule_sfs>:']
	return get_size_function(filename, subroutines)

def get_size_aes256_keyschedule_sfs_lut(filename):
	subroutines = ['<packing_rkey>:', '<aes256_keyschedule_rfunc_0>:',
		'<aes256_keyschedule_rfunc_1>:', '<inv_shiftrows_1>:',
		'<aes256_keyschedule_sfs_lut>:']
	return get_size_function(filename, subroutines)

def get_size_aes256_encrypt_sfs(filename):
	subroutines = ['<packing>:', '<unpacking>:', '<ark_sbox>:',
		'<mixcolumns_0>:', '<mixcolumns_3>:', '<double_shiftrows>:',
		'<aes256_encrypt_sfs>:']
	return get_size_function(filename, subroutines)

'''
AES-128 fully-fixsliced
'''
def get_size_aes128_keyschedule_ffs(filename):
	subroutines = ['<packing>:', '<aes128_xorcolumns_rotword>:',
		'<inv_shiftrows_1>:', '<loop_inv_sr_1>:', '<inv_shiftrows_2>:',
		'<loop_inv_sr_2>:', '<inv_shiftrows_3>:', '<loop_inv_sr_3>:',
		'<aes128_keyschedule_ffs>:']
	return get_size_function(filename, subroutines)

def get_size_aes128_keyschedule_ffs_lut(filename):
	subroutines = ['<packing_rkey>:', '<aes128_keyschedule_rfunc>:',
		'<inv_shiftrows_1>:', '<inv_shiftrows_2>:', '<inv_shiftrows_3>:',
		'<aes128_keyschedule_ffs_lut>:']
	return get_size_function(filename, subroutines)

def get_size_aes128_encrypt_ffs(filename):
	subroutines = ['<packing>:', '<unpacking>:', '<ark_sbox>:',
		'<mixcolumns_0>:', '<mixcolumns_1>:', '<mixcolumns_2>:',
		'<mixcolumns_3>:',  '<double_shiftrows>:', '<aes128_encrypt_ffs>:']
	return get_size_function(filename, subroutines)

'''
AES-256 fully-fixsliced
'''
def get_size_aes256_keyschedule_ffs(filename):
	subroutines = ['<packing>:', '<aes256_xorcolumns_rotword>:',
		'<aes256_xorcolumns>:', '<inv_shiftrows_1>:', '<loop_inv_sr_1>:',
		'<inv_shiftrows_2>:', '<loop_inv_sr_2>:', '<inv_shiftrows_3>:',
		'<loop_inv_sr_3>:', '<aes256_keyschedule_ffs>:']
	return get_size_function(filename, subroutines)

def get_size_aes256_keyschedule_ffs_lut(filename):
	subroutines = ['<packing_rkey>:', '<aes256_keyschedule_rfunc_0>:',
		'<aes256_keyschedule_rfunc_1>:', '<inv_shiftrows_1>:',
		'<inv_shiftrows_2>:', '<inv_shiftrows_3>:',
		'<aes256_keyschedule_ffs_lut>:']
	return get_size_function(filename, subroutines)

def get_size_aes256_encrypt_ffs(filename):
	subroutines = ['<packing>:', '<unpacking>:', '<ark_sbox>:',
		'<mixcolumns_0>:', '<mixcolumns_1>:', '<mixcolumns_2>:',
		'<mixcolumns_3>:',  '<double_shiftrows>:', '<aes256_encrypt_ffs>:']
	return get_size_function(filename, subroutines)

'''
AES-128 barrel-shiftrows
'''
def get_size_aes128_keyschedule_lut(filename):
	subroutines = ['<packing_rkey>:', '<packing_rkey_not>:',
		'<swapmove_rkey>:', '<keyschedule_round_func>:',
		'<aes128_keyschedule_lut>:', '<loop_aes128_keyschedule>:']
	return get_size_function(filename, subroutines)

def get_size_aes128_encrypt(filename):
	subroutines = ['<packing>:', '<unpacking>:', '<ark_sbox>:',
		'<add_round_key>:', '<sbox>:', '<shiftrows_1>:', '<shiftrows_2>:',
		'<shiftrows_3>:', '<mixcolumns>:', '<packing_0>:', '<packing_1>:', 
		'<packing_2>:', '<loop_p0>:', '<loop_p1>:', '<loop_p2>:',
		'<unpacking_0>:', '<unpacking_1>:', '<unpacking_2>:', '<loop_up0>:',
		'<loop_up1>:', '<loop_up2>:', '<aes128_encrypt>:',
		'<loop_aes128_core>:']
	return get_size_function(filename, subroutines)

'''
AES-256 barrel-shiftrows
'''
def get_size_aes256_keyschedule_lut(filename):
	subroutines = ['<packing_rkey>:', '<packing_rkey_not>:',
		'<swapmove_rkey>:', '<aes256_keyschedule_rfunc_0>:',
		'<aes256_keyschedule_rfunc_1>:', '<loop_aes256_keyschedule>:',
		'<aes256_keyschedule_lut>:']
	return get_size_function(filename, subroutines)

def get_size_aes256_encrypt(filename):
	subroutines = ['<packing>:', '<unpacking>:', '<ark_sbox>:',
		'<add_round_key>:', '<sbox>:', '<shiftrows_1>:', '<shiftrows_2>:',
		'<shiftrows_3>:', '<mixcolumns>:', '<packing_0>:', '<packing_1>:', 
		'<packing_2>:', '<loop_p0>:', '<loop_p1>:', '<loop_p2>:',
		'<unpacking_0>:', '<unpacking_1>:', '<unpacking_2>:', '<loop_up0>:',
		'<loop_up1>:', '<loop_up2>:', '<aes256_encrypt>:',
		'<loop_aes256_core>:']
	return get_size_function(filename, subroutines)

'''
Print the code size for all the AES encryption and keyschedule functions.
'''
def main(filename):
	# AES-128 semi-fixsliced
	print('--- AES-128 semi-fixsliced --- ')
	print('Code size for aes128_keyschedule_sfs is ' +
		str(get_size_aes128_keyschedule_sfs(filename)) + ' bytes.')
	print('Code size for aes128_keyschedule_sfs_lut is ' +
		str(get_size_aes128_keyschedule_sfs_lut(filename)) + ' bytes.')
	print('Code size for aes128_encrypt_sfs is ' +
		str(get_size_aes128_encrypt_sfs(filename)) + ' bytes.')
	# AES-256 semi-fixsliced
	print('--- AES-256 semi-fixsliced --- ')
	print('Code size for aes256_keyschedule_sfs is ' +
		str(get_size_aes256_keyschedule_sfs(filename)) + ' bytes.')
	print('Code size for aes256_keyschedule_sfs_lut is ' +
		str(get_size_aes256_keyschedule_sfs_lut(filename)) + ' bytes.')
	print('Code size for aes256_encrypt_sfs is ' +
		str(get_size_aes256_encrypt_sfs(filename)) + ' bytes.')
	# AES-128 fully-fixsliced
	print('--- AES-128 fully-fixsliced --- ')
	print('Code size for aes128_keyschedule_ffs is ' +
		str(get_size_aes128_keyschedule_ffs(filename)) + ' bytes.')
	print('Code size for aes128_keyschedule_ffs_lut is ' +
		str(get_size_aes128_keyschedule_ffs_lut(filename)) + ' bytes.')
	print('Code size for aes128_encrypt_ffs is ' +
		str(get_size_aes128_encrypt_ffs(filename)) + ' bytes.')
	# AES-256 fully-fixsliced
	print('--- AES-256 fully-fixsliced --- ')
	print('Code size for aes256_keyschedule_sfs is ' +
		str(get_size_aes256_keyschedule_ffs(filename)) + ' bytes.')
	print('Code size for aes256_keyschedule_sfs_lut is ' +
		str(get_size_aes256_keyschedule_ffs_lut(filename)) + ' bytes.')
	print('Code size for aes256_encrypt_ffs is ' +
		str(get_size_aes256_encrypt_ffs(filename)) + ' bytes.')
	# AES-128 barrel-shiftrows
	print('--- AES-128 barrel-shiftrows --- ')
	print('Code size for aes128_keyschedule_lut is ' +
		str(get_size_aes128_keyschedule_lut(filename)) + ' bytes.')
	print('Code size for aes128_encrypt is ' +
		str(get_size_aes128_encrypt(filename)) + ' bytes.')
	# AES-256 barrel-shiftrows
	print('--- AES-256 barrel-shiftrows --- ')
	print('Code size for aes256_keyschedule_lut is ' +
		str(get_size_aes256_keyschedule_lut(filename)) + ' bytes.')
	print('Code size for aes256_encrypt is ' +
		str(get_size_aes256_encrypt(filename)) + ' bytes.')

if __name__ == "__main__":
	if(len(sys.argv) == 1):
		print('Error: no input file.')
	else:
		main(sys.argv[1])
