'''
	Simple python3 script to get the code size of the AES encryption and 
	key schedule functions given at https://github.com/aadomn/aes.
	The output of riscv64-unknown-elf-objdump has to be given as input parameter.
	Put in the main program the functions that you want to be considered.

	author: Alexandre Adomnicai <alex.adomnicai@gmail.com>
	date: 	November 2020
'''
import sys

'''
Return the code size of a subroutine.
'''
def get_size_subroutine(filename, label):
	file = open(filename, 'r')
	start = 0
	end = 0
	line = file.readline()
	while line:
		if label in line: #as soon as we find the definition of the subroutine
			start = int(line.split()[0], base=16) #get its beginning address
			line = file.readline()
			while not(line == '\n'): #as soon as its definition ends
				line = file.readline()
			line = file.readline()
			end = int(line.split()[0], base=16) #get its ending address
			break
		line = file.readline()
	file.close()
	return end-start 

'''
Return the code size of a function, including the subroutines it uses.
'''
def get_size_function(filename, subroutines):
	total_size = 0
	for i in subroutines:
		total_size += get_size_subroutine(filename, i)
	return total_size

'''
AES-128 semi-fixsliced
'''
def get_size_aes128_keyschedule_sfs(filename):
	subroutines = ['<aes128_xorcolumns_rotword>:', '<inv_shiftrows_1>:',
		'<aes128_keyschedule_sfs>:']
	return get_size_function(filename, subroutines)

def get_size_aes128_keyschedule_sfs_lut(filename):
	subroutines = ['<subpacking_rkey>:', '<aes128_rfunc_ks>:',
		'<inv_shiftrows_1>:', '<store_not_rkey>:', '<aes128_keyschedule_sfs_lut>:',
		'<key_exp_loop_sfs>:']
	return get_size_function(filename, subroutines)

def get_size_aes128_encrypt_sfs(filename):
	subroutines = ['<packing_0>:', '<packing_1>:', '<packing_2>:',
		'<ark_sbox>:', '<mixcolumns0>:', '<mixcolumns3>:',
		'<double_shiftrows>:', '<aes128_encrypt_sfs>:']
	return get_size_function(filename, subroutines)

'''
AES-256 semi-fixsliced
'''
def get_size_aes256_keyschedule_sfs(filename):
	subroutines = ['<aes256_xorcolumns_rotword>:', '<inv_shiftrows_1>:',
		'<aes256_xorcolumns>:', '<aes256_keyschedule_sfs>:']
	return get_size_function(filename, subroutines)

def get_size_aes256_keyschedule_sfs_lut(filename):
	subroutines = ['<subpacking_rkey>:', '<aes256_rfunc_ks_0>:', 
		'<aes256_rfunc_ks_1>:', '<store_not_rkey>:',
		'<aes256_keyschedule_sfs_lut>:', '<inv_shiftrows_1>:',
		'<aes256_packing_loop_sfs_lut>:']
	return get_size_function(filename, subroutines)

def get_size_aes256_encrypt_sfs(filename):
	subroutines = ['<packing_0>:', '<packing_1>:', '<packing_2>:',
		'<ark_sbox>:', '<mixcolumns0>:', '<mixcolumns3>:',
		'<double_shiftrows>:', '<aes256_encrypt_sfs>:']
	return get_size_function(filename, subroutines)

'''
AES-128 fully-fixsliced
'''
def get_size_aes128_keyschedule_ffs(filename):
	subroutines = ['<aes128_xorcolumns_rotword>:', '<inv_shiftrows_1>:',
		'<double_shiftrows>:', '<inv_shiftrows_3>:',
		'<aes128_keyschedule_ffs>:']
	return get_size_function(filename, subroutines)

def get_size_aes128_keyschedule_ffs_lut(filename):
	subroutines = ['<subpacking_rkey>:', '<aes128_rfunc_ks>:',
		'<inv_shiftrows_1>:', '<inv_shiftrows_2>:', '<inv_shiftrows_3>:',
		'<store_not_rkey>:', '<aes128_keyschedule_sfs_lut>:',
		'<key_exp_loop_ffs>:']
	return get_size_function(filename, subroutines)

def get_size_aes128_encrypt_ffs(filename):
	subroutines = ['<packing_0>:', '<packing_1>:', '<packing_2>:',
		'<ark_sbox>:', '<mixcolumns0>:', '<mixcolumns1>:', '<mixcolumns3>:',
		'<double_shiftrows>:', '<aes128_encrypt_ffs>:']
	return get_size_function(filename, subroutines)

'''
AES-256 fully-fixsliced
'''
def get_size_aes256_keyschedule_ffs(filename):
	subroutines = ['<aes256_xorcolumns_rotword>:', '<inv_shiftrows_1>:',
		'<double_shiftrows>:', '<inv_shiftrows_3>:', '<aes256_xorcolumns>:',
		'<aes256_keyschedule_ffs>:']
	return get_size_function(filename, subroutines)

def get_size_aes256_keyschedule_ffs_lut(filename):
	subroutines = ['<subpacking_rkey>:', '<aes256_rfunc_ks_0>:', 
		'<aes256_rfunc_ks_1>:', '<store_not_rkey>:', '<aes256_keyschedule_ffs_lut:',
		'<inv_shiftrows_1>:', '<inv_shiftrows_2>:', '<inv_shiftrows_3>:',
		'<aes256_packing_loop_ffs>:']
	return get_size_function(filename, subroutines)

def get_size_aes256_encrypt_ffs(filename):
	subroutines = ['<packing_0>:', '<packing_1>:', '<packing_2>:',
		'<ark_sbox>:', '<mixcolumns0>:', '<mixcolumns1>:', '<mixcolumns3>:',
		'<double_shiftrows>:', '<aes256_encrypt_ffs>:']
	return get_size_function(filename, subroutines)

'''
AES-128 barrel-shiftrows
'''
def get_size_aes128_encrypt(filename):
	subroutines = ['<aes128_encrypt>:', '<packing_loop_0>:',
		'<packing_loop_1>:', '<packing_loop_2>:', '<unpacking_loop_0>:',
		'<unpacking_loop_1>:', '<unpacking_loop_2>:', '<unpacking_loop_2_bis>:',
		'<unpacking>:', '<aes128_addroundkey_sbox>:', '<aes128_linear_layer>:']
	return get_size_function(filename, subroutines)

def get_size_aes256_encrypt(filename):
	subroutines = ['<aes256_encrypt>:', '<aes256_packing_loop0>:',
		'<aes256_packing_loop1>:', '<aes256_packing_loop2>:',
		'<aes256_unpacking_loop0>:', '<aes256_unpacking_loop1>:',
		'<aes256_unpacking_loop2>:', '<aes256_unpacking_loop2_bis>:',
		'<aes256_unpacking>:', '<aes256_addroundkey_sbox>:',
		'<aes256_linear_layer>:']
	return get_size_function(filename, subroutines)
'''
Print the code size for all the AES encryption functions.
'''
def main(filename):
	# Fixslicing
	print('Code size for aes128_encrypt_sfs is ' +
		str(get_size_aes128_encrypt_sfs(filename)) + ' bytes.')
	print('Code size for aes256_encrypt_sfs is ' +
		str(get_size_aes256_encrypt_sfs(filename)) + ' bytes.')
	print('Code size for aes128_encrypt_ffs is ' +
		str(get_size_aes128_encrypt_ffs(filename)) + ' bytes.')
	print('Code size for aes256_encrypt_ffs is ' +
		str(get_size_aes256_encrypt_ffs(filename)) + ' bytes.')
	# # Barrel-shiftrows
	# print('Code size for aes128_encrypt is ' +
	# 	str(get_size_aes128_encrypt(filename)) + ' bytes.')
	# print('Code size for aes256_encrypt is ' +
	# 	str(get_size_aes256_encrypt(filename)) + ' bytes.')

if __name__ == "__main__":
	if(len(sys.argv) == 1):
		print('Error: no input file.')
	else:
		main(sys.argv[1])
